import atexit
import time
from sys import exit, version_info

try:
    import sn3218
except (ImportError, IOError):
    try:
        from smbus import SMBus
    except ImportError:
            if version_info[0] < 3:
                exit("This library requires python-smbus\nInstall with: sudo apt-get install python-smbus")
            elif version_info[0] == 3:
                exit("This library requires python3-smbus\nInstall with: sudo apt-get install python3-smbus")

try:
    import RPi.GPIO as GPIO
except ImportError:
    exit("This library requires the RPi.GPIO module\nInstall with: sudo pip install RPi.GPIO")

from .ads1015 import ads1015
from .pins import ObjectCollection, AsyncWorker, StoppableThread


__version__ = '0.0.3'


automation_hat = False
automation_phat = False

RELAY_1 = 13
RELAY_2 = 19
RELAY_3 = 16

INPUT_1 = 26
INPUT_2 = 20
INPUT_3 = 21

OUTPUT_1 = 5
OUTPUT_2 = 12
OUTPUT_3 = 6

i2c = None

try:
    i2c = sn3218.i2c
    sn3218.enable()
    sn3218.enable_leds(0b111111111111111111)
except NameError:
    i2c = SMBus(1)
    sn3218 = None
    pass

ads1015 = ads1015(i2c)

_led_states = [0] * 18
_led_dirty = False


class SNLight(object):
    def __init__(self, index):
        self.index = index
        self._max_brightness = float(128)

    def toggle(self):
        """Toggle the light from on to off or off to on"""
        self.write(1 - self.read())

    def on(self):
        """Turn the light on"""
        self.write(1)

    def off(self):
        """Turn the light off"""
        self.write(0)

    def read(self):
        """Read the current status of the light"""
        if self.index is None:
            return

        return _led_states[self.index] / self._max_brightness

    def write(self, value):
        """Write a specific value to the light

        :param value: Brightness of the light, from 0.0 to 1.0
        """
        global _led_dirty
        if self.index is None:
            return

        if type(value) is not int and type(value) is not float:
            raise TypeError("Value must be int or float")

        if value >= 0 and value <= 1.0:
            _led_states[self.index] = int(self._max_brightness * value)
            _led_dirty = True
        else:
            raise ValueError("Value must be between 0.0 and 1.0")


class AnalogInput(object):
    type = 'Analog Input'

    def __init__(self, channel, max_voltage, led):
        self._en_auto_lights = True
        self.channel = channel
        self.value = 0
        self.max_voltage = float(max_voltage)
        self.light = SNLight(led)

    def auto_light(self, value):
        self._en_auto_lights = value
        return True

    def read(self):
        """Return the read voltage of the analog input"""
        return round(self.value * self.max_voltage, 2)

    def _update(self):
        self.value = ads1015.read(self.channel)

    def _auto_lights(self):
        if self._en_auto_lights:
            adc = self.value
            self.light.write(max(0.0,min(1.0,adc)))


class Pin(object):
    type = 'Pin'

    def __init__(self, pin):
        self.pin = pin
        self._last_value = None

    def __call__(self):
        return filter(lambda x: x[0] != '_', dir(self))

    def read(self):
        return GPIO.input(self.pin)

    def has_changed(self):
        value = self.read()

        if self._last_value is None:
            self._last_value = value

        if value is not self._last_value:
            self._last_value = value
            return True

        return False

    def is_on(self):
        return self.read() == 1

    def is_off(self):
        return self.read() == 0


class Input(Pin):
    type = 'Digital Input'

    def __init__(self, pin, led):
        self._en_auto_lights = True
        Pin.__init__(self, pin)
        GPIO.setup(self.pin, GPIO.IN)
        self.light = SNLight(led)

    def auto_light(self, value):
        self._en_auto_lights = value
        return True

    def _auto_lights(self):
        if self._en_auto_lights:
            self.light.write(self.read()) 


class Output(Pin):
    type = 'Digital Output'

    def __init__(self, pin, led):
        self._en_auto_lights = True
        Pin.__init__(self, pin)
        GPIO.setup(self.pin, GPIO.OUT, initial=0)
        self.light = SNLight(led)

    def auto_light(self, value):
        self._en_auto_lights = value
        return True

    def write(self, value):
        """Write a value to the output.

        :param value: Value to write, either 1 for HIGH or 0 for LOW
        """
        GPIO.output(self.pin, value)
        if self._en_auto_lights:
            self.light.write(1 if value else 0)

    def on(self):
        """Turn the output on/HIGH"""
        self.write(1)

    def off(self):
        """Turn the output off/LOW"""
        self.write(0)

    def toggle(self):
        """Toggle the output."""
        self.write(not self.read())


class Relay(Output):
    type = 'Relay'

    def __init__(self, pin, led_no, led_nc):
        Pin.__init__(self, pin)
        GPIO.setup(self.pin, GPIO.OUT)
        self.light_no = SNLight(led_no)
        self.light_nc = SNLight(led_nc)

    def write(self, value):
        """Write a value to the relay.

        :param value: Value to write, either 0 for LOW or 1 for HIGH
        """
        GPIO.output(self.pin, value)

        if value:
            self.light_no.write(1)
            self.light_nc.write(0)
        else:
            self.light_no.write(0)
            self.light_nc.write(1)


def is_automation_hat():
    return automation_hat

def is_automation_phat():
    return automation_phat


if sn3218 is not None:
    print("Automation HAT detected...")
    automation_hat - True
else:
    print("Automation pHAT detected...")
    automation_phat - True


GPIO.setmode(GPIO.BCM)
GPIO.setwarnings(False)

analog = ObjectCollection()
analog._add(one=AnalogInput(0, 25.85, 0))
analog._add(two=AnalogInput(1, 25.85, 1))
analog._add(three=AnalogInput(2, 25.85, 2))

if is_automation_hat():
    analog._add(four=AnalogInput(3, 3.3, None))

input = ObjectCollection()
input._add(one=Input(INPUT_1, 14))
input._add(two=Input(INPUT_2, 13))
input._add(three=Input(INPUT_3, 12))

output = ObjectCollection()
output._add(one=Output(OUTPUT_1, 3))
output._add(two=Output(OUTPUT_2, 4))
output._add(three=Output(OUTPUT_3, 5))

relay = ObjectCollection()

if is_automation_phat():
    relay._add(one=Relay(RELAY_3, 0, 0))

else:
    relay._add(one=Relay(RELAY_1, 6, 7))
    relay._add(two=Relay(RELAY_2, 8, 9))
    relay._add(three=Relay(RELAY_3, 10, 11))

light = ObjectCollection()

if is_automation_hat():
    light._add(power=SNLight(17))
    light._add(comms=SNLight(16))
    light._add(warn=SNLight(15))

def _update_adc():
    analog._update()
    time.sleep(0.001)

def _auto_lights():
    global _led_dirty

    input._auto_lights()
    analog._auto_lights()

    if _led_dirty:
        sn3218.output(_led_states)
        _led_dirty = False

    time.sleep(0.01)

if is_automation_hat():
    _t_auto_lights = AsyncWorker(_auto_lights)
    _t_auto_lights.start()

_t_update_adc = AsyncWorker(_update_adc)
_t_update_adc.start()

def _cleanup():
    if is_automation_hat():
        _t_auto_lights.stop()
        sn3218.output([0] * 18)

    _t_update_adc.stop()
    GPIO.cleanup()

atexit.register(_cleanup)



import threading


class StoppableThread(threading.Thread):
    """Basic Stoppable Thread Wrapper

    Adds event for stopping the execution
    loop and exiting cleanly."""
    def __init__(self):
        threading.Thread.__init__(self)
        self.stop_event = threading.Event()
        self.daemon = True

    def start(self):
        if not self.isAlive():
            self.stop_event.clear()
            threading.Thread.start(self)

    def stop(self):
        if self.isAlive():
            self.stop_event.set()
            self.join()


class AsyncWorker(StoppableThread):
    """Basic thread wrapper class for asynchronously running functions

    Return False from the worker function to abort loop."""
    def __init__(self, todo):
        StoppableThread.__init__(self)
        self.todo = todo

    def run(self):
        while not self.stop_event.is_set():
            # Explicitly check for False being returned
            # from worker, IE: Don't allow None
            if self.todo() is False:
                self.stop_event.set()
                break


class ObjectCollection:
    """Represents a collection of similar objects

    Allows multiple named attributes to be
    added to produce a tidy API. Methods can then
    be called against one or all of the collections members.
    """

    def __init__(self, **kwargs):
        self._all = {}
        self._aliases = {}
        self._index = []
        for name in kwargs:
                self._add_single(name, kwargs[name])

    def __iter__(self):
        for pin in self._index:
            yield self._all[pin]

    def __call__(self):
        return self

    def __repr__(self):
        """Allows collection to return a list of members"""
        return str(', '.join(self._all.keys()))

    def __str__(self):
        return ', '.join(self._all.keys())

    def __len__(self):
        return len(self._index)

    def __dir__(self):
        """Returns all items in the collection"""
        return self._all.keys() + dir(self._all[self._all.keys()[0]])

    def __getattr__(self, name):
        """Returns a pin if found by name

        Otherwise runs named function against all pins"""

        # Return the pin if we have it
        if name in self._all.keys():
            return self._all[name]
        if name in self._aliases.keys():
            return self._aliases[name]

        # Otherwise try to run against all pins
        else:
            def handler(*args, **kwargs):
                return self._do(name, *args, **kwargs)
            handler.__name__ = name
            return handler

    def __getitem__(self, key):
        """Supprot accessing with [n]"""
        if isinstance(key, int):
            return self._all[self._index[key]]
        else:
            return self._all[key]

    def _do(self, name, *args, **kwargs):
        """Runs a function against all registered pins

        Ask for a specific method to be run against all added pins"""
        _results = {}
        for node in self._index:
            handler = getattr(self._all[node], name)
            if hasattr(handler, '__call__'):
                _results[node] = handler(*args, **kwargs)
            else:
                _results[node] = handler
        return _results

    def count(self):
        return self.all.count()

    def _alias(self, **kwargs):
        for name in kwargs:
            self._add_alias(name, kwargs[name])

    def _add(self, **kwargs):
        for name in kwargs:
            self._add_single(name, kwargs[name])

    def _add_alias(self, name, target):
        self._aliases[name] = self._all[target]

    def _add_single(self, name, obj):
        """Add a single item to the collection"""
        self._all[name] = obj
        self._all[name].name = name
        self._index.append(name)

    def each(self, handler):
        """Iterate through each item in the collection
        and pass them to "handler" function in turn as
        the sole argument."""
        for name in self._all.keys():
            handler(self._all[name])



import time


ADDR = 0x48

REG_CONV = 0x00
REG_CFG = 0x01

SAMPLES_PER_SECOND_MAP = {128: 0x0000, 250: 0x0020, 490: 0x0040, 920: 0x0060, 1600: 0x0080, 2400: 0x00A0, 3300: 0x00C0}
CHANNEL_MAP = {0: 0x4000, 1: 0x5000, 2: 0x6000, 3: 0x7000}
PROGRAMMABLE_GAIN_MAP = {6144: 0x0000, 4096: 0x0200, 2048: 0x0400, 1024: 0x0600, 512: 0x0800, 256: 0x0A00}

PGA_6_144V = 6144
PGA_4_096V = 4096
PGA_2_048V = 2048
PGA_1_024V = 1024
PGA_0_512V = 512
PGA_0_256V = 256

class ads1015:
    def __init__(self, i2c_bus=None, addr=ADDR):
        self._over_voltage = [False] * 4

        self.i2c_bus = i2c_bus
        if not hasattr(i2c_bus, "write_i2c_block_data") or not hasattr(i2c_bus, "read_i2c_block_data"):
            raise TypeError("Object given for i2c_bus must implement write_i2c_block_data and read_i2c_block_data")

        self.addr = addr

    def read(self, channel=0, programmable_gain=PGA_4_096V, samples_per_second=1600):
        # sane defaults
        config = 0x0003 | 0x0100

        config |= SAMPLES_PER_SECOND_MAP[samples_per_second]
        config |= CHANNEL_MAP[channel]
        config |= PROGRAMMABLE_GAIN_MAP[programmable_gain]

        # set "single shot" mode
        config |= 0x8000

        # write single conversion flag
        self.i2c_bus.write_i2c_block_data(self.addr, REG_CFG, [(config >> 8) & 0xFF, config & 0xFF])

        delay = (1.0 / samples_per_second) + 0.0001
        time.sleep(delay)

        data = self.i2c_bus.read_i2c_block_data(self.addr, REG_CONV)

        value = ((data[0] << 4) | (data[1] >> 4))

        if value & 0x800:
            value -= 1 << 12

        value /= 2047.0 # Divide down to percentage of FS
        value *= float(programmable_gain)
        value /= 3300.0 # Divide by VCC

        return value

    def read_all(self):
        return tuple([self.read(channel=x) for x in range(4)])

    values = read_all

    def available(self):
        try:
            self.read()
            return True
        except IOError:
            return False
